@@ -4,6 +4,7 @@
 # imports.
 from ssht00ls.v1.classes.config import *
 from ssht00ls.v1.classes import utils
+from ssht00ls.v1.classes.smart_cards import smart_cards
 
 # the installation object class.
 class Installation(object):
@@ -554,7 +555,7 @@ class SSH(object):
 		port=22,
 		key_path=None,
 	):
-		base = __create_base__()
+		base = []
 		if alias != None:
 			return base + ['ssh', alias]
 		else:
@@ -587,7 +588,7 @@ class SSHFS(object):
 	):
 
 		# checks.
-		base = __create_base__()
+		base = []
 		if alias == None:
 			success, response = utils.__check_parameters__(empty_value=None, parameters={
 				"username":username,
@@ -610,8 +611,8 @@ class SSHFS(object):
 			alias = f'{username}@{ip}'
 
 		# do.
-		output = utils.__execute__(base + [f'{alias}:{server_path}', client_path], timeout=3)
-		#output = utils.__execute_script__(utils.__array_to_string__(base + [f'{alias}:{server_path}', client_path], joiner="\n"), timeout=3)
+		output = utils.__execute__(base + [f'{alias}:{server_path}', client_path])
+		#output = utils.__execute_script__(utils.__array_to_string__(base + [f'{alias}:{server_path}', client_path], joiner="\n"))
 
 		# check fails.
 		if "mount_osxfuse: mount point " in output and "is itself" in output:
@@ -820,9 +821,9 @@ class SCP(object):
 		port=22,
 		key_path=None,
 	):	
-		base = __create_base__()
+		base = []
 		if alias != None:
-			return base + ['ssh', alias]
+			return base + ['scp']
 		else:
 			base += ['scp']
 			if key_path != None: base += ['-i', key_path,]
@@ -856,8 +857,10 @@ class Agent(object):
 		smart_card=False, 
 		# the smart cards pin code
 		pin=None, 
-		# default timeout (do not edit).
+		# default timeout (do not use).
 		timeout=0.5,
+		# reattempt (do not use).
+		reattempt=False,
 	):
 
 		# initialize.
@@ -880,22 +883,32 @@ class Agent(object):
 		# with passphrase.
 		response = utils.__default_response__()
 		if smart_card or passphrase != None:
-			self.initialize()
+			#self.initialize()
 			if smart_card:
+				path = smart_cards.path
 				if OS in ["osx"]:
-					copy_path = "/usr/local/lib/libykcs11_NOTALINK.dylib" # working version 1
-					os.system(f"rm -fr {copy_path}")
-					os.system(f"cp {path} {copy_path}")
-					path = copy_path
-				file = Files.File(path="/tmp/shell.sh")
+					os.system(f"rm -fr {smart_cards.path}")
+					os.system(f"cp {smart_cards.original_path} {smart_cards.path}")
+					os.system(f"chmod 644 {smart_cards.path}")
+				#file = Files.File(path="/tmp/shell.sh")
 				#file.save(f'export SSH_AUTH_SOCK={self.SSH_AUTH_SOCK}\nexport SSH_AGENT_PID={self.SSH_AGENT_PID}\nssh-add -e {path}\nssh-add -s {path}')
-				file.save(f'ssh-add -e {path}\nssh-add -s {path}')
+				#file.save(f'ssh-add -e {path}\nssh-add -s {path}')
 				#file.save(f'ssh-add -s {path}')
-				file.file_path.permission.set(755)
-				spawn = pexpect.spawn(f'sh {file.file_path.path}')
-				file.file_path.delete(forced=True)
-				#spawn = pexpect.spawn(f'ssh-add {path}')
+				#file.file_path.permission.set(755)
+				#spawn = pexpect.spawn(f'sh {file.file_path.path}')
+				#file.file_path.delete(forced=True)
+				#spawn = pexpect.spawn(f'ssh-add -e {path} && ssh-add -s {path}')
+				try:
+					output = subprocess.check_output([f"ssh-add", "-e", f"{path}"])
+					print("remove card output:",output)
+				except: a=1
+				spawn = pexpect.spawn(f'ssh-add -s {path}')
 			else:
+				#file = Files.File(path="/tmp/shell.sh")
+				#file.save(f'export SSH_AUTH_SOCK={self.SSH_AUTH_SOCK}\nexport SSH_AGENT_PID={self.SSH_AGENT_PID}\nssh-add {path}')
+				#file.file_path.permission.set(755)
+				#spawn = pexpect.spawn(f'sh {file.file_path.path}')
+				#file.file_path.delete(forced=True)
 				spawn = pexpect.spawn(f'ssh-add {path}')
 
 			# send lines.
@@ -932,7 +945,35 @@ class Agent(object):
 			elif "incorrect pin" in output.lower():
 				return r3sponse.error_response("Provided an incorrect pin code.")
 			elif "Failed to communicate" in output or "agent refused operation" in output:
-				return r3sponse.error_response("Failed to communicate with the ssh-agent. Try logging out the current system user & logging back in (or execute [$ eval `ssh-agent`]).")
+				if reattempt:
+					try:
+						output = subprocess.check_output([f"ssh-add", "-D"])
+						print("delete output:",output)
+					except: a=1
+					try:
+						output = subprocess.check_output([f"ssh-add", "-k"])
+						print("kill output:",output)
+					except: a=1
+					try:
+						output = subprocess.check_output([f"eval", "`ssh-agent`"])
+						print("restart output:",output)
+					except: a=1
+					return self.add(
+						# the keys path.
+						path=path,
+						# the keys passphrase.
+						passphrase=passphrase,
+						# enable if you are using a smart card.
+						smart_card=smart_card,
+						# the smart cards pin code
+						pin=pin,
+						# default timeout (do not use).
+						timeout=timeout,
+						# reattempt (do not use).
+						reattempt=False,
+					)
+				else:
+					return r3sponse.error_response("Failed to communicate with the ssh-agent. Try logging out the current system user & logging back in (or execute [$ eval `ssh-agent`]).")
 			elif "Identity added:" in output or "Card added:" in output: 
 				return r3sponse.success_response(f"Successfully added key [{path}] to the ssh agent.")
 			elif output != "": 
@@ -1078,27 +1119,332 @@ class Agent(object):
 		response["message"] = f"Successfully initialized the ssh agent."
 		return response
 
-def __create_base__():
-	return []
-	path = "/tmp/shell_source.sh"
-	utils.__save_file__(path, f"#!/bin/sh\nSSH_AUTH_SOCK={agent.SSH_AUTH_SOCK}; export SSH_AUTH_SOCK;\nSSH_AGENT_PID={agent.SSH_AGENT_PID}; export SSH_AGENT_PID\n")
-	utils.__set_file_path_permission__(path, permission=755)
-	base = [
-		f"{path}", "&&",
-	]
-	return base
+# the sshd object class.
+class SSHD(object):
+	def __init__(self):
+		a=1
+	def create(self,
+		# save the configuration & banner.
+		save=False,
+		# the ssh port.
+		port=22,
+		# the listen addresses.
+		listen_addresses=[],
+		# the server's banner.
+		banner="Hello World!",
+		# the allowed users & options.
+		users={
+			# define per user (all keys are optional).
+			"administrator": {
+				# the user's root permissions.
+				"root_permissions":False,
+				# authentication by password.
+				"password_authentication":False,
+				# authentication by keys.
+				"key_authentication":True,
+				# ip filter.
+				"ip_filter":False,
+				"allowed_ips":[],
+				# sftp server only.
+				"sftp_only":False,
+				# the chroot directory (leave null to disable).
+				"chroot_directory":None,
+				# allowed connection options.
+				"x11_forwarding":False,
+				"tcp_forwarding":False,
+			},
+		},
+	):
 
+		# check users.
+		response = self.__check_user_items__(users)
+		if response["error"] != None: return response
+
+		# check utils intalled (must be before __install_banner__).
+		response = self.__check_utils_installed__(list(users.keys()))
+		if response["error"] != None: return response
+
+		# intall banner.
+		if save:
+			response = self.__install_banner__(banner=banner, usernames=list(users.keys()))
+			if response["error"] != None: return response
+
+		# defaults.
+		configuration =  '# SSHD_CONFIG:'
+		configuration += '\n# BY VANDENBERGHINC'
+		configuration += '\n# MODULE: ssht00ls'
+		configuration += '\n# AUTHOR: DAAN VAN DEN BERGH'
+		configuration += '\nAcceptEnv LANG LC_*'
+		configuration += '\nSubsystem sftp  internal-sftp'
+		#configuration += '\nSubsystem sftp  /usr/libexec/sftp-server'
+		configuration += '\nLoginGraceTime 60'
+		configuration += '\nMaxAuthTries 3'
+		configuration += '\nMaxSessions 10'
+		configuration += "\nMaxStartups 999"
+		configuration += '\nLogLevel VERBOSE'
+		configuration += f'\nPort {port}'
+		configuration += '\nProtocol 2'
+
+		# defaults.
+		configuration += '\nPermitRootLogin {}'.format("no")
+		configuration += '\nStrictModes {}'.format("yes")
+		configuration += '\nPermitUserEnvironment {}'.format("no")
+		configuration += '\nIgnoreRhosts {}'.format("yes")
+		configuration += '\nPermitTunnel {}'.format("no")
+		configuration += '\nX11Forwarding {}'.format("no")
+		configuration += '\nAllowTcpForwarding {}'.format("no")
+		configuration += '\nAllowStreamLocalForwarding {}'.format("no")
+		configuration += '\nGatewayPorts {}'.format("no")
+		configuration += '\nPermitTTY {}'.format("yes")
+		for listen_address in listen_addresses:
+			configuration += f'\nListenAddress {listen_address}'
+
+		# auth keys.
+		configuration += '\nAuthorizedKeysFile {}'.format(".ssh/authorized_keys")
+
+		# per users.
+		for username, info in users.items():
+			configuration += f'\n# User: {username}'
+
+			# ip filter.	
+			configuration += f'\nMatch User {username}'
+
+			# authentication by password.
+			if info["password_authentication"]:
+				configuration += '\n     PasswordAuthentication yes'
+				configuration += '\n     PermitEmptyPasswords no'
+				configuration += '\n     ChallengeResponseAuthentication no'
+			else:
+				configuration += '\n     PasswordAuthentication no'
+				configuration += '\n     PermitEmptyPasswords no'
+				configuration += '\n     ChallengeResponseAuthentication no'
+
+			# authentication by keys.
+			if info["key_authentication"]:
+				configuration += '\n     PubkeyAuthentication {}'.format('yes')
+			else:
+				configuration += '\n     PubkeyAuthentication {}'.format('no')
+
+
+			# chroot directory.
+			if isinstance(info["chroot_directory"], str):
+				configuration += f'\n     ChrootDirectory {info["chroot_directory"]}'
+
+			# root permission.
+			l = "no"
+			if info["root_permissions"] and info["key_authentication"]: l = "prohibit-password"
+			elif info["root_permissions"]: l = "yes"
+			configuration += f'\n     PermitRootLogin {l}'
+
+			# connection options.
+			configuration += f'\n     X11Forwarding {self.__convert_boolean__(info["x11_forwarding"])}'
+			configuration += f'\n     AllowTcpForwarding {self.__convert_boolean__(info["tcp_forwarding"])}'
+
+			# default options.
+			configuration += f'\n     PermitTunnel no'
+			configuration += f'\n     AllowStreamLocalForwarding no'
+			configuration += f'\n     GatewayPorts no'
+			configuration += f'\n     PermitTTY yes'
+			
+			# check ip filter.
+			if info["ip_filter"]:
+
+				# match verified ips.
+				configuration += f'\n     Match User {username} Address {self.__sum_list__(info["allowed_ips"])}'
+
+				# check sftp only.
+				if info["sftp_only"]:
+					configuration += '\n          ForceCommand internal-sftp'
+
+				# shell access.
+				else:
+					configuration += '\n          ForceCommand .ssh/utils/original.sh'
+
+				# match unverified ips.
+				configuration += f'\n     Match User {username} Address *,!{self.__sum_list__(info["allowed_ips"])}'
+				configuration += f'\n          ForceCommand .ssh/utils/log.sh "Your ip address is not authorized." "Authorize your ip address to access user [{username}]."'
+
+			# no ip filter.
+			else:
+
+				# check sftp only.
+				if info["sftp_only"]:
+					configuration += '\n          ForceCommand internal-sftp'
+
+				# shell access.
+				else:
+					configuration += '\n          ForceCommand .ssh/utils/original.sh'
+
+		# match none authorized users.
+		#if '*all*' not in list(users.keys()):
+		configuration += f'\nMatch User *,!{self.__sum_list__(list(users.keys()))}'
+		configuration += f'\n     ForceCommand .ssh/utils/log.sh "You are not authorized to access user [$USER] over ssh."'
+		configuration += "\n"
+
+		# save sshd.
+		if save:
+			file = Files.File(path='/tmp/sshd_config', data=configuration)
+			file.file_path.delete(forced=True, sudo=True)
+			file.save()
+			fp = Formats.FilePath(f"/etc/ssh/sshd_config")
+			file.file_path.copy(fp.path, sudo=True)
+			fp.permission.set(permission=644, sudo=True)
+			fp.ownership.set(owner="root", group=None, sudo=True)
+			if not fp.exists(sudo=True):
+				return r3sponse.error_response(f"Failed to save the sshd configuration.")
+
+		# success.
+		return r3sponse.success_response("Successfully created the sshd configuration.", {
+				"sshd":configuration,
+			})
+
+		#
+	# system functions.
+	def __sum_list__(self, list):
+		return Files.Array(path=False, array=list).string(joiner=',')
+	def __convert_boolean__(self, boolean):
+		if boolean: return "yes"
+		else: return "no"
+	def __check_user_items__(self, users):
+
+		# iterate.
+		for username, info in users.items():
+			
+			# check options.
+			try: info["root_permissions"]
+			except KeyError: info["root_permissions"] = True
+			try: info["password_authentication"]
+			except KeyError: info["password_authentication"] = False
+			try: info["key_authentication"]
+			except KeyError: info["key_authentication"] = True
+			try: info["ip_filter"]
+			except KeyError: info["ip_filter"] = False
+			try: 
+				info["allowed_ips"]
+				if not isinstance(info["allowed_ips"], list):
+					return r3sponse.error_response(f"Invalid usage, parameter [users.{username}.allowed_ips] is supposed to be a list with allowed ip addresses.")
+			except KeyError: info["allowed_ips"] = []
+			try: info["sftp_only"]
+			except KeyError: info["sftp_only"] = False
+			try: info["chroot_directory"]
+			except KeyError: info["chroot_directory"] = None
+			try: info["x11_forwarding"]
+			except KeyError: info["x11_forwarding"] = False
+			try: info["tcp_forwarding"]
+			except KeyError: info["tcp_forwarding"] = False
+
+		# response.
+		return r3sponse.success_response("Successfully checked the user items.")
+
+		#
+	def __check_utils_installed__(self, usernames=[]):
+
+		# iterate.
+		if isinstance(usernames, str): usernames = [usernames]
+		to_install = []
+		for username in usernames:
+			
+			# non existant.
+			fp = Formats.FilePath(f"{HOME_BASE}{username}/.ssh/utils/.version.py")
+			if not fp.exists(sudo=True): 
+				to_install.append(username)
+
+			# check version.
+			else: 
+				version = utils.__execute__(["sudo", "cat", fp.path])
+				if str(version) != str(Files.File(path=f'{SOURCE_PATH}/{VERSION}/classes/utils/.version.py', load=True).data):
+					to_install.append(username)
+
+		# install.
+		if len(to_install) > 0:
+			response = self.__install_utils__(to_install)
+			if response["error"] != None: return response
+
+		# success.
+		return r3sponse.success_response("Successfully verified the ssht00ls utils installation.")
+
+		#
+	def __install_utils__(self, usernames=[]):
+
+		# checks.
+		if isinstance(usernames, str): usernames = [usernames]
+		if len(usernames) == 0: 
+			return r3sponse.error_response("No usernames specified.")
+
+		# download.
+		path = "/tmp/utils/"
+		os.system(f"rm -fr {path}")
+		os.system(f"rm -fr {HOME}/Downloads/ssht00ls/")
+		os.system(f'git clone https://github.com/vandenberghinc/ssht00ls {HOME}/Downloads/ssht00ls/')
+		if not os.path.exists(f"{HOME}/Downloads/ssht00ls/"):
+			return r3sponse.error_response("Failed to install the ssht00ls utils. #1")
+		os.system(f"mv {HOME}/Downloads/ssht00ls/ssht00ls/v1/classes/utils/ {path}")
+		os.system(f"rm -fr {path}/__pycache__")
+		os.system(f"rm -fr {path}/__init__.py")
+		if not os.path.exists(path):
+			return r3sponse.error_response("Failed to install the ssht00ls utils. #2")
+
+		# iterate.
+		for username in usernames:
+
+			# check if ssh is correctly installed.
+			response = installation.check_installed(username=username)
+
+			# install the ssh correctly for the specified user.
+			if response["error"] != None:
+				response = installation.install(username=username)
+				if response["error"] != None: return response
+
+			# copy.
+			fp = Formats.FilePath(f"{HOME_BASE}{username}/.ssh/utils/")
+			fp.delete(forced=True, sudo=True)
+			os.system(f"sudo cp -r {path} {fp.path}")
+			fp.ownership.set(owner=username, group=None, sudo=True)
+			fp.permission.set(permission=755, recursive=True, sudo=True)
+			if not fp.exists(sudo=True):
+				return r3sponse.error_response("Failed to install the ssht00ls utils. #3")
+
+		# success.
+		return r3sponse.success_response("Successfully installed the ssht00ls utils.")
+
+		#
+	def __install_banner__(self, banner="", usernames=[]):
+
+		# checks.
+		if isinstance(usernames, str): usernames = [usernames]
+		if len(usernames) == 0: 
+			return r3sponse.error_response("No usernames specified.")
+
+		# save banner.
+		file = Files.File(path='/tmp/banner', data=banner)
+		file.file_path.delete(forced=True, sudo=True)
+		file.save()
+
+		# iterate.
+		for username in usernames:
+			fp = Formats.FilePath(f"/{HOME_BASE}{username}/.ssh/banner")
+			file.file_path.copy(fp.path, sudo=True)
+			fp.permission.set(permission=755, sudo=True)
+			fp.ownership.set(owner=username, group=None, sudo=True)
+			if not fp.exists(sudo=True):
+				return r3sponse.error_response(f"Failed to install the banner for user [{username}].")
+
+		# success.
+		return r3sponse.success_response("Successfully installed the banner.")
+
+		#
+	#
 
 # Initialized classes.
 installation = Installation()
 config = Config()
+sshd = SSHD()
 ssh = SSH()
 sshfs = SSHFS()
 scp = SCP()
 agent = Agent()
 key = Key()
 
-
 """
 
 # --------------------
@@ -1112,6 +1458,42 @@ response = installation.check_installed(username=None)
 if response["error"] != None:
 	response = installation.install(username=None)
 
+# --------------------
+# SSHD Config.
+
+# create and optionally save a sshd configuration.
+response = sshd.create(
+	# save the configuration & banner.
+	save=False,
+	# the ssh port.
+	port=22,
+	# the listen addresses.
+	listen_addresses=[],
+	# the server's banner.
+	banner="Hello World!",
+	# the allowed users & options.
+	users={
+		# define per user (all keys are optional).
+		"administrator": {
+			# the user's root permissions.
+			"root_permissions":False,
+			# authentication by password.
+			"password_authentication":False,
+			# authentication by keys.
+			"key_authentication":True,
+			# ip filter.
+			"ip_filter":True,
+			"allowed_ips":["192.168.1.201"],
+			# sftp server only.
+			"sftp_only":False,
+			# the chroot directory (leave null to disable).
+			"chroot_directory":None,
+			# allowed connection options.
+			"x11_forwarding":False,
+			"tcp_forwarding":False,
+		},
+	},)
+
 # --------------------
 # SSH Key.
 
